# -*- coding: utf-8 -*-
from . import STDR

class MTR(STDR):
     def __init__(self, version=None, endian=None, record=None,  BSR__ADDR_SIZ=None, BSR__WC_SIZ=None):
         self.id = 'MTR'
         self.local_debug = False
         if version == 'V4':
             self.version = version
             self.info=    '''
 Memory Test Record (V4, Memory:2010.1)
 --------------------------------------

 Function:
     This is the record is used to store fail data along with capture test conditions and references to test test descriptions.
     It allows the fail data to be stored in various formats describe below using the field highlighting

 Frequency:
     Number of memory tests times records required to log the fails for the test (counting continuation record)

 Location:
     It can occur after all the memory design specific records i.e. any Memory Structure Record (MSR),
     any Memory Controller Records (MCRs), any Memory Instance Records (IDRs), any Memory Model Records(MMRs),
     any Algorithms Specification Records (ASRs), any Frame Specification Records (FSRs) and any Bitstream Specificaion Records (BSRs)
 '''
             #TODO: Implement "Field Presense Expression" (see PTR record on how)
             self.fields = {
                 'REC_LEN'   : {'#' :  0, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' :    None},
                 'REC_TYP'   : {'#' :  1, 'Type' :  'U*1', 'Ref' :                     None, 'Value' :   15, 'Text' : 'Record type                           ', 'Missing' :    None},
                 'REC_SUB'   : {'#' :  2, 'Type' :  'U*1', 'Ref' :                     None, 'Value' :   40, 'Text' : 'Record sub-type                       ', 'Missing' :    None},
                 'CONT_FLG'  : {'#' :  3, 'Type' :  'B*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Continuation flag                     ', 'Missing' :    None},
                 'TEST_NUM'  : {'#' :  4, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Test number                           ', 'Missing' :    None},
                 'HEAD_NUM'  : {'#' :  5, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Test head number                      ', 'Missing' :       1},
                 'SITE_NUM'  : {'#' :  6, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Test site number                      ', 'Missing' :       1},
                 'ASR_REF'   : {'#' :  7, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'ASR Index                             ', 'Missing' :    None},
                 'TEST_FLG'  : {'#' :  8, 'Type' :  'B*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Test flags (fail, alarm, etc.)        ', 'Missing' : ['0']*8},
                 'LOG_TYP'   : {'#' :  9, 'Type' :  'C*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'User defined description of datalog   ', 'Missing' :      ''},
                 'TEST_TXT'  : {'#' : 10, 'Type' :  'C*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Descriptive text or label             ', 'Missing' :      ''},
                 'ALARM_ID'  : {'#' : 11, 'Type' :  'C*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Name of alarm                         ', 'Missing' :      ''},
                 'PROG_TXT'  : {'#' : 12, 'Type' :  'C*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Additional Programmed information     ', 'Missing' :      ''},
                 'RSLT_TXT'  : {'#' : 13, 'Type' :  'C*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Additional result information         ', 'Missing' :      ''},
                 'COND_CNT'  : {'#' : 14, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (k) of conditions               ', 'Missing' :       0},
                 'COND_LST'  : {'#' : 15, 'Type' : 'xC*n', 'Ref' :               'COND_CNT', 'Value' : None, 'Text' : 'Array of Conditions                   ', 'Missing' :      []},
                 'CYC_CNT'   : {'#' : 16, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Total cycles executed during the test ', 'Missing' :       0},
                 'TOTF_CNT'  : {'#' : 17, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Total fails during the test           ', 'Missing' :       0},
                 'TOTL_CNT'  : {'#' : 18, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Total fails during the complete MTR   ', 'Missing' :       0},
                 'OVFL_FLG'  : {'#' : 19, 'Type' :  'B*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Failure Flag                          ', 'Missing' : ['0']*8},
                 'FILE_INC'  : {'#' : 20, 'Type' :  'B*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'File incomplete                       ', 'Missing' : ['0']*8},
                 'LOG_TYPE'  : {'#' : 21, 'Type' :  'B*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Type of datalog                       ', 'Missing' : ['0']*8},
                 'FDIM_CNT'  : {'#' : 22, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (m) of FDIM_FNAM and FDIM_FCNT  ', 'Missing' :       0},
                 'FDIM_NAM'  : {'#' : 23, 'Type' : 'xC*n', 'Ref' :               'FDIM_CNT', 'Value' : None, 'Text' : 'Array of logged Dim names             ', 'Missing' :      []},
                 'FDIM_FCNT' : {'#' : 24, 'Type' : 'xU*8', 'Ref' :               'FDIM_CNT', 'Value' : None, 'Text' : 'Array of failure counts               ', 'Missing' :      []},
                 'CYC_BASE'  : {'#' : 25, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Cycle offset to CYC_OFST array        ', 'Missing' :       0},
                 'CYC_SIZE'  : {'#' : 26, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Size (f) of CYC_OFST [1,2,4 or 8 byes]', 'Missing' :       1},
                 'PMR_SIZE'  : {'#' : 27, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Size (f) of PMR_ARR [1 or 2 bytes]    ', 'Missing' :       1},
                 'ROW_SIZE'  : {'#' : 28, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Size (f) of ROW_ARR [1,2,4 or 8 bytes]', 'Missing' :       1},
                 'COL_SIZE'  : {'#' : 29, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Size (f) of COL_ARR [1,2,4 or 8 bytes]', 'Missing' :       1},
                 'DLOG_MSK'  : {'#' : 30, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Presence indication mask              ', 'Missing' :       0},
                 'PMR_CNT'   : {'#' : 31, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (n) of pins in PMN_ARR          ', 'Missing' :       0},
                 'PMR_ARR'   : {'#' : 32, 'Type' : 'xU*f', 'Ref' :  ('PMR_CNT', 'PMR_SIZE'), 'Value' : None, 'Text' : 'Array of PMR indexes for pins         ', 'Missing' :      []},
                 'CYCO_CNT'  : {'#' : 33, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (n) of CYC_OFST array           ', 'Missing' :       0},
                 'CYC_OFST'  : {'#' : 34, 'Type' : 'xU*f', 'Ref' : ('CYCO_CNT', 'CYC_SIZE'), 'Value' : None, 'Text' : 'Array of cycle indexes for each fail  ', 'Missing' :      []},
                 'ROW_CNT'   : {'#' : 35, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (d) of ROW_ARR array            ', 'Missing' :       0},
                 'ROW_ARR'   : {'#' : 36, 'Type' : 'xU*f', 'Ref' :  ('ROW_CNT', 'ROW_SIZE'), 'Value' : None, 'Text' : 'Array of row addresses for each fail  ', 'Missing' :      []},
                 'COL_CNT'   : {'#' : 37, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (d) of COL_ARR array            ', 'Missing' :       0},
                 'COL_ARR'   : {'#' : 38, 'Type' : 'xU*f', 'Ref' :  ('COL_CNT', 'COL_SIZE'), 'Value' : None, 'Text' : 'Array of col addresses for each fail  ', 'Missing' :      []},
                 'STEP_CNT'  : {'#' : 39, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (d) STEP_ARR array              ', 'Missing' :       0},
                 'STEP_ARR'  : {'#' : 40, 'Type' : 'xU*1', 'Ref' :               'STEP_CNT', 'Value' : None, 'Text' : 'Array of march steps for each fail    ', 'Missing' :      []},
                 'DIM_CNT'   : {'#' : 41, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Number (k) of dimensions              ', 'Missing' :       0},
                 'DIM_NAMS'  : {'#' : 42, 'Type' : 'xC*n', 'Ref' :                'DIM_CNT', 'Value' : None, 'Text' : 'Names of the dimensions               ', 'Missing' :      []},
                 'DIM_DCNT'  : {'#' : 43, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (n) of DIM_VALS                 ', 'Missing' :       0},
                 'DIM_DSIZ'  : {'#' : 44, 'Type' :  'U*1', 'Ref' :                     None, 'Value' : None, 'Text' : 'Size (f) of DIM_VALS [1,2,4or 8 bytes]', 'Missing' :       1},
                 'DIM_VALS'  : {'#' : 45, 'Type' : 'xU*f', 'Ref' : ('DIM_DCNT', 'DIM_DSIZ'), 'Value' : None, 'Text' : 'Array of data values for a dimension  ', 'Missing' :      []},
                 'TFRM_CNT'  : {'#' : 46, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Total frames in frame based logging   ', 'Missing' :       0},
                 'TFSG_CNT'  : {'#' : 47, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Total segments across all records     ', 'Missing' :       0},
                 'LFSG_CNT'  : {'#' : 48, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'Local number of frame segments        ', 'Missing' :       0},
                 'FRM_IDX'   : {'#' : 49, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'Index of the frame record             ', 'Missing' :       0},
                 'FRM_MASK'  : {'#' : 50, 'Type' :  'D*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Frame presence mask                   ', 'Missing' :      []},
                 'FRM_CNT'   : {'#' : 51, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count (q) of frame (curr frame & maks)', 'Missing' :       0},
                 'LFBT_CNT'  : {'#' : 52, 'Type' :  'U*4', 'Ref' :                     None, 'Value' : None, 'Text' : 'Count(q) of bits stored in this record', 'Missing' :       0},
                 'FRAMES'    : {'#' : 53, 'Type' :  'D*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Bit encoded data (curr FSR)           ', 'Missing' :      []},
                 'TBSG_CNT'  : {'#' : 54, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'Number of logged bit stream segments  ', 'Missing' :       0},
                 'LBSG_CNT'  : {'#' : 55, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : '# of bit stream segmnts in this record', 'Missing' :       0},
                 'BSR_IDX'   : {'#' : 56, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'Index of the bit stream record        ', 'Missing' :       0},
                 'STRT_ADR'  : {'#' : 57, 'Type' :  'U*f', 'Ref' :            BSR__ADDR_SIZ, 'Value' : None, 'Text' : 'Start row addr in the current segment ', 'Missing' :       1},
                 'WORD_CNT'  : {'#' : 58, 'Type' :  'U*f', 'Ref' :              BSR__WC_SIZ, 'Value' : None, 'Text' : 'Word count in current stream segment  ', 'Missing' :       1},
                 'WORDS'     : {'#' : 59, 'Type' :  'D*n', 'Ref' :                     None, 'Value' : None, 'Text' : 'Bit encoded data for one or words     ', 'Missing' :      []},
                 'TBMP_SIZE' : {'#' : 60, 'Type' :  'U*8', 'Ref' :                     None, 'Value' : None, 'Text' : 'count (k) of CBIT_MAP                 ', 'Missing' :       0},
                 'LBMP_SIZE' : {'#' : 61, 'Type' :  'U*2', 'Ref' :                     None, 'Value' : None, 'Text' : 'Bytes from map in the current record  ', 'Missing' :       0},
                 'CBIT_MAP'  : {'#' : 62, 'Type' : 'xU*1', 'Ref' :              'TBMP_SIZE', 'Value' : None, 'Text' : 'Compressed bit map                    ', 'Missing' :      []}
             }
         else:
             raise STDR.STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
         self._default_init(endian, record)



# -*- coding: utf-8 -*-
from . import STDR

class CDR(STDR):
     def __init__(self, version=None, endian=None, record = None):
         self.id = 'CDR'
         self.local_debug = False
         if version==None or version=='V4':
             self.version = 'V4'
             self.info=    '''
# Chain Description Record (V4-2007)
# ----------------------------------

# Function:
#     This record contains the description of a scan chain in terms of its input, output, number of cell and clocks.
#     Each CDR record contains description of exactly one scan chain. Each CDR is uniquely identified by an index.

# Frequency:
#     ?!?

# Location:
#     ?!?
# '''
             self.fields = {
                 'REC_LEN'  : {'#' :  0, 'Type' : 'U*2',  'Ref' : None,       'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                 'REC_TYP'  : {'#' :  1, 'Type' : 'U*1',  'Ref' : None,       'Value' :    1, 'Text' : 'Record type                           ', 'Missing' : None},
                 'REC_SUB'  : {'#' :  2, 'Type' : 'U*1',  'Ref' : None,       'Value' :   94, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                 'CONT_FLG' : {'#' :  3, 'Type' : 'B*1',  'Ref' : None,       'Value' : None, 'Text' : 'Continuation CDR record follow (if!=0)', 'Missing' : 0},
                 'CDR_INDX' : {'#' :  4, 'Type' : 'U*2',  'Ref' : None,       'Value' : None, 'Text' : 'SCR Index                             ', 'Missing' : 0},
                 'CHN_NAM'  : {'#' :  5, 'Type' : 'C*n',  'Ref' : None,       'Value' : None, 'Text' : 'Chain Name                            ', 'Missing' : None},
                 'CHN_LEN'  : {'#' :  6, 'Type' : 'U*4',  'Ref' : None,       'Value' : None, 'Text' : 'Chain Length (cells in chain)         ', 'Missing' : 0},
                 'SIN_PIN'  : {'#' :  7, 'Type' : 'U*2',  'Ref' : None,       'Value' : None, 'Text' : "PMR index of the chain's Scan In Sig  ", 'Missing' : 0},
                 'SOUT_PIN' : {'#' :  8, 'Type' : 'U*2',  'Ref' : None,       'Value' : None, 'Text' : "PMR index of the chain's Scan Out Sig ", 'Missing' : 0},
                 'MSTR_CNT' : {'#' :  9, 'Type' : 'U*1',  'Ref' : None,       'Value' : None, 'Text' : 'Count (m) of master clock pins        ', 'Missing' : 0},
                 'M_CLKS'   : {'#' : 10, 'Type' : 'xU*2', 'Ref' : 'MSTR_CNT', 'Value' : None, 'Text' : 'Arr of PMR indses for the master clks ', 'Missing' : []},
                 'SLAV_CNT' : {'#' : 11, 'Type' : 'U*1',  'Ref' : None,       'Value' : None, 'Text' : 'Count (n) of slave clock pins         ', 'Missing' : 0},
                 'S_CLKS'   : {'#' : 12, 'Type' : 'xU*2', 'Ref' : 'SLAV_CNT', 'Value' : None, 'Text' : 'Arr of PMR indxes for the slave clks  ', 'Missing' : []},
                 'INV_VAL'  : {'#' : 13, 'Type' : 'U*1',  'Ref' : None,       'Value' : None, 'Text' : '0: No Inversion, 1: Inversion         ', 'Missing' : 0},
                 'LST_CNT'  : {'#' : 14, 'Type' : 'U*2',  'Ref' : None,       'Value' : None, 'Text' : 'Count (k) of scan cells               ', 'Missing' : 0},
                 'CELL_LST' : {'#' : 15, 'Type' : 'xS*n', 'Ref' : 'LST_CNT',  'Value' : None, 'Text' : 'Array of Scan Cell Names              ', 'Missing' : []},
             }
         else:
             raise STDR.STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
         self._default_init(endian, record)



'''
Created on 11 Aug 2019

@author: tho
'''
import os
import re
import struct
import io

from . import (ATR, BPS, CDR, CNR, DTR, EPS, FAR, FTR, GDR, HBR, MIR, MPR, MRR, 
               NMR, PCR, PGR, PIR, PLR, PMR, PRR, PSR, PTR, RDR, SBR, SDR, SSR, 
               STR, TSR, VUR, WIR, WCR, WRR)

from Semi_ATE.STDF.STDR import STDFError
from Semi_ATE.STDF.STDR import get_STDF_setup_from_file
from Semi_ATE.STDF.STDR import get_bytes_from_file
from Semi_ATE.STDF.STDR import ts_to_id
from Semi_ATE.STDF.STDR import id_to_ts
from Semi_ATE.STDF.STDR import supported

# def stdfopen(FileName, mode='rb'):
#     '''
#     returns an open file object to the stdf file.
#     if the file is no STDF file, None is returned.
#     if the STDF file is compressed, returns the file object of the correct algorithm.

#     mode = 'rb' or 'wb' (no text supported ... raise ValueError if not rb or wb)
#     '''
#     if mode!='rb' and mode!='rb': raise ValueError("Only 'rb' and 'wb' are supported.")
#     if not is_STDF(FileName): return None



# def to_df(FileName, progress=True):
#     '''
#     This function will return a pandas data-frame from the given FileName.

#     This process has 3 stages :
#         1) index the FileName
#         2) Analyse
#         2) construct the dataframe

#     ---> needs to move to metis !!! metis.import_stdf(...)

#     '''

#     index = {}
# #   index = {'version' : stdf_version,
# #            'endian'  : stdf_endian,
# #            'records' : { REC_NAM : [offset, ...
# #            'indexes' : { offset : bytearray of the record ...
# #            'parts' : {part# : [offset, offset, offset, ...
#     offset = 0

#     if is_STDF(FileName):
#         endian, version = endian_and_version_from_file(FileName)
#         index['version'] = version
#         index['endian'] = endian
#         index['records'] = {}
#         index['indexes'] = {}
#         index['parts'] = {}
#         PIP = {} # parts in process
#         PN = 1

#         TS2ID = ts_to_id(version)

#         if progress:
#             description = "Indexing STDF file '%s'" % os.path.split(FileName)[1]
#             index_progress = tqdm(total=get_deflated_file_size(FileName), ascii=True, disable=not progress, desc=description, leave=False, unit='b')

#         for _, REC_TYP, REC_SUB, REC in records_from_file(FileName):
#             REC_ID = TS2ID[(REC_TYP, REC_SUB)]
#             REC_LEN = len(REC)
#             if REC_ID not in index['records']: index['records'][REC_ID] = []
#             index['indexes'][offset] = REC
#             index['records'][REC_ID].append(offset)
#             if REC_ID in ['PIR', 'PRR', 'PTR', 'FTR', 'MPR']:
#                 if REC_ID == 'PIR':
#                     pir = PIR(index['version'], index['endian'], REC)
#                     pir_HEAD_NUM = pir.get_value('HEAD_NUM')
#                     pir_SITE_NUM = pir.get_value('SITE_NUM')
#                     if (pir_HEAD_NUM, pir_SITE_NUM) in PIP:
#                         raise Exception("One should not be able to reach this point !")
#                     PIP[(pir_HEAD_NUM, pir_SITE_NUM)] = PN
#                     index['parts'][PN]=[]
#                     index['parts'][PN].append(offset)
#                     PN+=1
#                 elif REC_ID == 'PRR':
#                     prr = PRR(index['version'], index['endian'], REC)
#                     prr_HEAD_NUM = prr.get_value('HEAD_NUM')
#                     prr_SITE_NUM = prr.get_value('SITE_NUM')
#                     if (prr_HEAD_NUM, prr_SITE_NUM) not in PIP:
#                         raise Exception("One should not be able to reach this point!")
#                     pn = PIP[(prr_HEAD_NUM, prr_SITE_NUM)]
#                     index['parts'][pn].append(offset)
#                     del PIP[(prr_HEAD_NUM, prr_SITE_NUM)]
#                 elif REC_ID == 'PTR':
#                     ptr = PTR(index['version'], index['endian'], REC)
#                     ptr_HEAD_NUM = ptr.get_value('HEAD_NUM')
#                     ptr_SITE_NUM = ptr.get_value('SITE_NUM')
#                     if (ptr_HEAD_NUM, ptr_SITE_NUM) not in PIP:
#                         raise Exception("One should not be able to reach this point!")
#                     pn = PIP[(ptr_HEAD_NUM, ptr_SITE_NUM)]
#                     index['parts'][pn].append(offset)
#                 elif REC_ID == 'FTR':
#                     ftr = FTR(index['version'], index['endian'], REC)
#                     ftr_HEAD_NUM = ftr.get_value('HEAD_NUM')
#                     ftr_SITE_NUM = ftr.get_value('SITE_NUM')
#                     if (ftr_HEAD_NUM, ftr_SITE_NUM) not in PIP:
#                         raise Exception("One should not be able to reach this point!")
#                     pn = PIP[(ftr_HEAD_NUM, ftr_SITE_NUM)]
#                     index['parts'][pn].append(offset)
#                 elif REC_ID == 'MPR':
#                     mpr = MPR(index['version'], index['endian'], REC)
#                     mpr_HEAD_NUM = mpr.get_value('HEAD_NUM')
#                     mpr_SITE_NUM = mpr.get_value('SITE_NUM')
#                     if (mpr_HEAD_NUM, mpr_SITE_NUM) not in PIP:
#                         raise Exception("One should not be able to reach this point!")
#                     pn = PIP[(mpr_HEAD_NUM, mpr_SITE_NUM)]
#                     index['parts'][pn].append(offset)
#                 else:
#                     raise Exception("One should not be able to reach this point! (%s)" % REC_ID)
#             if progress: index_progress.update(REC_LEN)
#             offset += REC_LEN

#         if progress:
#             description = "Analyzing data"
#             ttl = len(index['records']['TSR'])
#             analyze_progress = tqdm(total=ttl, ascii=True, position=1, disable=not progress, desc=description, leave=False, unit='tests')

#         TEST_NUM_NAM = {}

#         for tsr_offset in index['records']['TSR']:
#             tsr = TSR(index['version'], index['endian'], index['indexes'][tsr_offset])
#             TEST_NUM = tsr.get_value('TEST_NUM')
#             TEST_NAM = tsr.get_value('TEST_NAM')
#             TEST_TYP = tsr.get_value('TEST_TYP').upper()
#             if TEST_NUM not in TEST_NUM_NAM:
#                 TEST_NUM_NAM[TEST_NUM] = []
#             if (TEST_NAM, TEST_TYP) not in TEST_NUM_NAM[TEST_NUM]:
#                 TEST_NUM_NAM[TEST_NUM].append((TEST_NAM, TEST_TYP))
#             analyze_progress.update()

#         for TEST_NUM in TEST_NUM_NAM:
#             if len(TEST_NUM_NAM[TEST_NUM])==1:
#                 TEST_NUM_NAM[TEST_NUM] = TEST_NUM_NAM[TEST_NUM][0]


#         # Create the indexes of the dataframe
#         ROW_index = sorted(list(index['parts']))
#         TEST_ITM_index = ['LOT_ID', 'MOD_COD', 'X_POS', 'Y_POS'] #TODO: add more ...
#         TEST_NAM_index = ['Meta'] * len(TEST_ITM_index)
#         TEST_NUM_index = ['Meta'] * len(TEST_ITM_index)
#         for TEST_NUM in sorted(TEST_NUM_NAM):
#             TEST_TYP = TEST_NUM_NAM[TEST_NUM][1]
#             if TEST_TYP == 'P':
#                 PTR_FIELDS = ['LO_SPEC', 'LO_LIMIT', 'RESULT', 'HI_LIMIT', 'HI_LIMIT', 'UNITS', 'PF']
#                 TEST_ITM_index+=PTR_FIELDS
#                 TEST_NAM_index+=[TEST_NUM_NAM[TEST_NUM][1]]*len(PTR_FIELDS)
#                 TEST_NUM_index+=[TEST_NUM]*len(PTR_FIELDS)
#             elif TEST_TYP == 'F':

#                 TEST_NUM_index+=[TEST_NUM]*5
#                 TEST_NAM_index+=[TEST_NUM_NAM[TEST_NUM][1]]*5     # VECT_NAME TIME_SET NUM_FAIL X_FAIL_AD Y_FAIL_AD PF
#             elif TEST_TYP == 'M':
#                 pass
#             else:
#                 raise STDFError("Test Type '%s' is unknown" % TEST_TYP)




#         print("\n\n\n")


#         for record_offset in index['parts'][1]:
#             record = index['indexes'][record_offset]
#             T, S = TS_from_record(record)
#             ID = TS2ID[(T,S)]
#             if ID == 'PTR':
#                 ptr = PTR(index['version'], index['endian'], record)
#                 print(ptr)
#             if ID == 'PIR':
#                 pir = PIR(index['version'], index['endian'], record)
#                 print(pir)
#             if ID == 'PRR':
#                 prr = PRR(index['version'], index['endian'], record)
#                 print(prr)





# #         if progress:
# #             description = "Constructing data-frame"
# #             constructing_progress = tqdm(total=len(index['parts']), ascii=True, position=2, disable=not progress, desc=description, leave=False, unit='parts')
# #
# #         for part in index['parts']:
# #             for record_offset in index['parts'][part]:
# #                 Type, SubType = TS_from_record(index['indexes'][record_offset])
# #                 ID = TS2ID[(Type, SubType)]
# #                 if ID == 'FTR':
# #                     ftr = FTR(index['version'], index['endian'], index['indexes'][record_offset])
# #
# #
# #                 if ID == 'PTR':
# #                     ptr = PTR(index['version'], index['endian'], index['indexes'][record_offset])
# #
# #                 if ID == 'MPR':
# #                     ptr = MPR(index['version'], index['endian'], index['indexes'][record_offset])
# #
# #
# #
# #             constructing_progress.update()

#         if progress:
#             index_progress.close()
#             analyze_progress.close()
# #             constructing_progress.close()

#         return index, TEST_NUM_NAM

#     else: #not an STDF file
#         pass

supported_compressions = {'lzma' : '.xz', 'gzip' : '.gz', 'bz2' : '.bz2'}
supported_compressions_extensions = {supported_compressions[k]:k for k in supported_compressions}


def os_is_case_sensitive():
    if os.path.normcase('A') == os.path.normcase('a'):
        return False
    return True


def has_valid_STDF_extension(FileName):
    '''
    returns True if according to the standard, this is a valid STDF extension
    '''
    if os_is_case_sensitive():
        regex = r'std'
    else:
        regex = r'STD'
    elements = re.split(regex, FileName)
    if len(elements) == 1:
        return False
    return True

def has_pretty_STDF_extension(FileName):
    '''
    looks if the uncompressed file name ends with '.std' or '.stdf'
    looks if the (supported) compressed file name ends with '.std.xx' or '.stdf.xx'
    here xx are the supported compression extensions
    '''
    if not has_valid_STDF_extension(FileName):
        return False  # based on filename
    if not is_STDF(FileName):
        return False  # based on contents
    if is_compressed_file(FileName, list(supported_compressions_extensions)):
        ext = extension_from_magic_number_in_file(FileName)
        if len(ext) == 1 and (FileName.upper().endswith(".STD%s" % ext[0].upper()) or FileName.upper().endswith('.STDF%s' % ext[0].upper())):
            return True
    else:
        if FileName.upper().endswith(".STD") or FileName.upper().endswith('.STDF'):
            return True
    return False

def set_pretty_STDF_extension(FileName, use_hash=False):
    '''

    TODO: implement hashing
    '''


def is_WS(FileName, progress=False):
    '''
    This function returns True if the given (compressed) FileName is made during Wafer Sort, False otherwise.
    The only reliable way to determine if the data is generated during Wafer Sort is to look for the presense of 'WIR'.
    This function might take a while on big files, so if progress=True, a progress bar is displayed.
    '''
    raise NotImplementedError("Woops, maybe now is a good moment to implement")
    return True

def is_FT(FileName, progress=False):
    '''
    This function returns True if the given (compressed) FileName is made during Final Test, False otherwhise.
    The only reliable way to determine if the date is generated dureing Final Test is to look for the absense of 'WIR'.
    This function might take a while on big files, so if progress=True, a progress bar is displayed.
    '''
    return not is_WS(FileName, progress)

def is_STDF(FileName):
    '''
    This function will read the first 4 bytes of a file, and see if byte 3 == 0 and byte 4 ==10
    (that is the magic number of an STDF file) if so return True, False otherwise.

    Note, it is checked if the file is compressed (only supports gzip, bz2 and lzma), if so,
    the uncompressed file is examined.
    '''
    if not os.path.exists(FileName):
        return False

    if not os.path.isfile(FileName):
        return False

    if is_compressed_file(FileName, ['.gz', '.xz', '.bz2']):
        extension = extension_from_magic_number_in_file(FileName)[0]
        if extension == '.gz':
            import gzip
            with gzip.open(FileName, 'rb') as fd:
                FAR = fd.read(4)
                REC_TYP, REC_SUB = struct.unpack('BB', FAR[2:4])
                if REC_TYP == 0 and REC_SUB == 10:
                    return True
                else:
                    return False
        elif extension == '.bz2':
            import bz2
            with bz2.open(FileName, 'rb') as fd:
                FAR = fd.read(4)
                REC_TYP, REC_SUB = struct.unpack('BB', FAR[2:4])
                if REC_TYP == 0 and REC_SUB == 10:
                    return True
                else:
                    return False
        elif extension == '.xz':
            import lzma
            with lzma.open(FileName, 'rb') as fd:
                FAR = fd.read(4)
                REC_TYP, REC_SUB = struct.unpack('BB', FAR[2:4])
                if REC_TYP == 0 and REC_SUB == 10:
                    return True
                else:
                    return False
        else:
            raise Exception("Shouldn't reach this point (%s)" % extension)
    else:
        with open(FileName, 'rb') as fd:
            FAR = fd.read(4)
            REC_TYP, REC_SUB = struct.unpack('BB', FAR[2:4])
            if REC_TYP == 0 and REC_SUB == 10:
                return True
            else:
                return False

def is_supported_compressed_STDF_file(FileName):
    '''
    Returns True if FileName is a supported compressed file, False otherwise
    '''
    if not is_STDF(FileName):
        return False
    ext = extension_from_magic_number_in_file(
        FileName, supported_compressions_extensions)
    if len(ext) != 1:
        return False
    return True

def endian_and_version_from_file(FileName):
    '''
    Returns the endian and version from FileName.
    if something went wrong, both are empty strings.
    '''
    tmp = records_from_file(FileName)
    endian = ''
    version = ''
    if tmp != None:  # success
        endian = tmp.endian
        version = tmp.version
    return endian, version

def MIR_from_file(FileName):
    '''
    return *THE* MIR object from FileName.
    '''
    endian, version = endian_and_version_from_file(FileName)
    if endian == '' or version == '':
        return MIR()
    for _, REC_TYP, REC_SUB, REC in records_from_file(FileName):
        if REC_TYP == 1 and REC_SUB == 10:
            break
    return MIR(version, endian, REC)

def SDRs_from_file(FileName):
    '''
    return the SDR(s) objects from FileName.
    '''
    retval = []
    endian, version = endian_and_version_from_file(FileName)
    print(endian, version)
    if endian == '' or version == '':
        return retval
    for _, REC_TYP, REC_SUB, REC in records_from_file(FileName):
        if (REC_TYP, REC_SUB) not in [(0, 10), (0, 20), (1, 10), (1, 70), (1, 80)]:
            break
        if (REC_TYP, REC_SUB) == (1, 80):
            retval.append(REC)
    return retval

def TS_from_record(record):
    '''
    given an STDF record (bytearray), extract the REC_TYP and REC_SUB
    Note: This will work on *ALL* records.
    '''
    return struct.unpack("BB", record[2:4])

def HEAD_NUM_and_SITE_NUM_from_record(record):
    '''
    given and STDF record (bytearray), extract the HEAD_NUM and SITE_NUM
    Note : HEAD_NUM and SITE_NUM are not always located at the same (byte) offset.

                REC_TYP   REC_SUB   HEAD_NUM   SITE_NUM (V4)
           PCR      1        30        4          5
           HBR      1        40        4          5
           SBR      1        50        4          5
           PMR      1        60     variable   variable <-- will not support here
           SDR      1        80        4        array   <-- will not support here
           WIR      2        10        4          /     <-- will not support here
           WRR      2        20        4          /     <-- will not support here
           PIR      5        10        4          5
           PRR      5        20        4          5
           TSR      10       30        4          5
           PTR      15       10        9          10    < most likely
           MPR      15       15        9          10    < most likely
           FTR      15       20        9          10    < most likely
    '''
    REC_TYP, REC_SUB = TS_from_record(record)
    HEAD_NUM = 0
    SITE_NUM = 0
    if REC_TYP == 15:  # PTR, MTR & FTR
        if REC_SUB in [10, 15, 20]:
            HEAD_NUM, SITE_NUM = struct.unpack("BB", record[8:10])
    elif REC_TYP == 5:  # PIR, PRR
        if REC_SUB in [10, 20, 30]:
            HEAD_NUM, SITE_NUM = struct.unpack("BB", record[4:6])
    elif REC_TYP == 1:  # PCR, HBR, SBR
        if REC_SUB in [30, 40, 50]:
            HEAD_NUM, SITE_NUM = struct.unpack("BB", record[4:6])
    return HEAD_NUM, SITE_NUM

def TEST_NUM_from_record(record, endian):
    '''
    given a PTR, MPR or FTR record, extract the test Number.
    Note: TEST_NUM is for these records always located on offset 4..7

                REC_TYP   REC_SUB   TEST_NUM (V4)
           PTR      15       10       4:7
           MPR      15       15       4:7
           FTR      15       20       4:7

          Also note that endian is important here!
    '''
    REC_TYP, REC_SUB = TS_from_record(record)
    TEST_NUM = -1
    if REC_TYP == 15 and REC_SUB in [10, 15, 20]:
        TEST_NUM = struct.unpack("%sI" % endian, record[4:7])
    return TEST_NUM

compression_extensions = ['.gz', '.7z', '.zip', '.xz', '.bz2']

def is_compressed_file(FileName, extensions_of_interest=compression_extensions):
    '''
    This function returns True if it is determined that FileName is compressed, False otherwise.
    Note: it will use the extension_from_magic_number function of this module.
    '''
    ext = extension_from_magic_number_in_file(FileName)
    if len(ext)==1: # a compressed file is unambiguous
        if ext[0] in extensions_of_interest:
            return True
    return False


class check_records_from_file(object):
    '''
    This is a *QUICK* iterator class that returns the next record from an STDF file each time it is called.
    It is fast because it doesn't check versions, extensions and it doesn't unpack the record and skips unknown records.
    It does support gzip, bz2 and lzma compression.
    '''
    def __init__(self, FileName):
        self.fd = None
        if not isinstance(FileName, str):
            return
        if not os.path.exists(FileName):
            return
        if not os.path.isfile(FileName):
            return
        if not is_STDF(FileName):
            return
        if is_supported_compressed_STDF_file(FileName):
            ext = extension_from_magic_number_in_file(FileName)
            if len(ext) != 1:
                return
            compression = supported_compressions_extensions[ext[0]]
            if compression == 'lzma':
                import lzma
                self.fd = lzma.open(FileName, 'rb')
            elif compression == 'bz2':
                import bz2
                self.fd = bz2.open(FileName, 'rb')
            elif compression == 'gzip':
                import gzip
                self.fd = gzip.open(FileName, 'rb')
            else:
                raise Exception(
                    "the %s compression is supported but not fully implemented." % compression)
        else:
            self.fd = open(FileName, 'rb')
        buff = self.fd.read(6)
        CPU_TYPE, STDF_VER = struct.unpack('BB', buff[4:])
        if CPU_TYPE == 1:
            self.endian = '>'
        elif CPU_TYPE == 2:
            self.endian = '<'
        else:
            self.endian = '?'
        self.version = 'V%s' % STDF_VER
        self.fd.seek(0)
        self.unpack_fmt = '%sHBB' % self.endian

    def __del__(self):
        if self.fd != None:
            self.fd.close()

    def __iter__(self):
        return self

    def __next__(self):
        while self.fd != None:
            while True:
                header = self.fd.read(4)
                if len(header) != 4:
                    raise StopIteration
                REC_LEN, REC_TYP, REC_SUB = struct.unpack(
                    self.unpack_fmt, header)
                footer = self.fd.read(REC_LEN)
                if len(footer) != REC_LEN:
                    raise StopIteration
                return REC_LEN, REC_TYP, REC_SUB, header+footer

def object_from_json(json):
    '''
    ToDo

    Parameters
    ----------
    object : json string

    Returns
    -------
    STDR objects like MIR, FAR etc..

    '''
extensions = {'.gz'   : [[(0, b'\x1f\x8b\x08')]], # gzip
              '.pdf'  : [[(0, b'\x25\x50\x44\x46\x2d')]],
              '.wav'  : [[(0, b'\x52\x49\x46\x46'), (8, b'\x57\x41\x56\x45')]],
              '.avi'  : [[(0, b'\x52\x49\x46\x46'), (8, b'\x41\x56\x49\x20')]],
              '.mp3'  : [[(0, b'\xFF\xFB')],
                         [(0, b'\x49\x44\x33')]],
              '.stdf' : [[(2, b'\x00\x0A')]],
              '.rpm'  : [[(0, b'\xed\xab\xee\xdb')]],
              '.ico'  : [[(0, b'\x00\x00\x01\x00')]],
              '.z'    : [[(0, b'\x1F\x9D')],
                         [(0, b'1F A0')]],
              '.bz2'  : [[(0, b'\x42\x5A\x68')]],
              '.gif'  : [[(0, b'\x47\x49\x46\x38\x37\x61')],
                         [(0, b'\x47\x49\x46\x38\x39\x61')]],
              '.tiff' : [[(0, b'\x49\x49\x2A\x00')],
                         [(0, b'\x4D\x4D\x00\x2A')]],
              '.exr'  : [[(0, b'\x76\x2F\x31\x01')]],
              '.bpg'  : [[(0, b'\x42\x50\x47\xFB')]],
              '.jpg'  : [[(0, b'\xFF\xD8\xFF\xDB')],
                         [(0, b'\xFF\xD8\xFF\xE0\x00\x10\x4A\x46\x49\x46\x00\x01')],
                         [(0, b'\xFF\xD8\xFF\xEE')],
                         [(0, b'\xFF\xD8\xFF\xE1'), (6, b'\x45\x78\x69\x66\x00\x00')]],
              '.lz'   : [[(0, b'\x4C\x5A\x49\x50')]],
              '.xls'  : [[(0, b'\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1')]],
              '.zip'  : [[(0, b'\x50\x4B\x03\x04')],
                         [(0, b'\x50\x4B\x05\x06')],
                         [(0, b'\x50\x4B\x07\x08')]],
              '.rar'  : [[(0, b'\x52\x61\x72\x21\x1A\x07\x00')],
                         [(0, b'\x52\x61\x72\x21\x1A\x07\x01\x00')]],
              '.png'  : [[(0, b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A')]],
              '.ps'   : [[(0, b'\x25\x21\x50\x53')]],
              '.ogg'  : [[(0, b'\x4F\x67\x67\x53')]],
              '.psd'  : [[(0, b'\x38\x42\x50\x53')]],
              '.bmp'  : [[(0, b'\x42\x4D')]],
              '.iso'  : [[(0, b'\x43\x44\x30\x30\x31')]],
              '.flac' : [[(0, b'\x66\x4C\x61\x43')]],
              '.midi' : [[(0, b'\x4D\x54\x68\x64')]],
              '.vmdk' : [[(0, b'\x4B\x44\x4D')]],
              '.dmg'  : [[(0, b'\x78\x01\x73\x0D\x62\x62\x60')]],
              '.xar'  : [[(0, b'\x78\x61\x72\x21')]],
              '.tar'  : [[(0, b'\x75\x73\x74\x61\x72\x00\x30\x30')],
                         [(0, b'\x75\x73\x74\x61\x72\x20\x20\x00')]],
              '.7z'   : [[(0, b'\x37\x7A\xBC\xAF\x27\x1C')]], # 7-Zip
              '.xz'   : [[(0, b'\xFD\x37\x7A\x58\x5A\x00\x00')]], # lzma
              '.XML'  : [[(0, b'\x3c\x3f\x78\x6d\x6c\x20')]],
              '.swf'  : [[(0, b'\x43\x57\x53')],
                         [(0, b'\x46\x57\x53')]],
              '.deb'  : [[(0, b'\x21\x3C\x61\x72\x63\x68\x3E')]],
              '.rtf'  : [[(0, b'\x7B\x5C\x72\x74\x66\x31')]],
              '.xcf'  : [[(0, b'\x67\x69\x6d\x70\x20\x78\x63\x66\x20')]],
              '.xlsx' : [[(0, b'\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1')], #password protected
                         [(0, b'\x50\x4B\x03\x04\x14\x00\x06\x00')]], # not password protected
              '.docx' : [[(0, b'\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1')], #password protected
                         [(0, b'\x50\x4B\x03\x04\x14\x00\x06\x00')]], # not password protected
              '.pptx' : [[(0, b'\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1')], #password protected
                         [(0, b'\x50\x4B\x03\x04\x14\x00\x06\x00')]], # not password protected

              }    
known_extensions = [ext for ext in extensions]

def extension_from_magic_number_in_file(FileName, extensions_of_interest=known_extensions):
    '''
    This function will try to determine the type of 'FileName' by looking at it's contents.
    returns the supposed extension (with the '.') of the fileType or None if nothing is recognized.
    Note: it doesn't look at the extension of a filename like 'mimetypes' does !
    Ref: https://en.wikipedia.org/wiki/List_of_file_signatures
    '''
    debug = False
    prv = []
    if debug: print(FileName)
    if os.path.exists(FileName) and os.path.isfile(FileName):
        with open(FileName, 'rb') as fd:
            for extension in extensions:
                possibilities = len(extensions[extension])
                if debug: print("\t'%s'" % extension)
                for possibility, magic_number in enumerate(extensions[extension]):
                    parts = len(magic_number)
                    if debug: print("\t\tPossibility %s/%s has %s parts: '%s'" % (possibility+1, possibilities, parts, magic_number))
                    tmp = []
                    for part, definition in enumerate(magic_number):
                        offset, pattern = definition
                        if debug: print("\t\t\t part=%s/%s : offset = %d, magic = %s, lenght = %d bytes" % (part+1, parts, definition[0], definition[1], len(pattern)), end = '')
                        fd.seek(offset)
                        data = fd.read(len(pattern))
                        if debug: print(" --> %s == %s ? " % (data, pattern), end='')
                        if data == pattern:
                            tmp.append(extension)
                            if debug: print('YES')
                        else:
                            if debug: print('NO')
                    if len(tmp) == parts:
                        prv.append(extension)
    retval = []
    for ext in prv:
        if ext in extensions_of_interest:
            retval.append(ext)
    return retval

def objects_from_indexed_file(FileName, index, records_of_interest=None):
    '''
     This is a Generator of records (not in order!)
    '''
    if not isinstance(FileName, str): raise STDFError("'%s' is not a string.")
    if not os.path.exists(FileName): raise STDFError("'%s' does not exist")
    endian = get_STDF_setup_from_file(FileName)[0]
    RLF = '%sH' % endian
    version = 'V%s' % struct.unpack('B', get_bytes_from_file(FileName, 5, 1))
    fd = open(FileName, 'rb')

    ALL = list(id_to_ts(version).keys())
    if records_of_interest==None:
        roi = ALL
    elif isinstance(records_of_interest, list):
        roi = []
        for item in records_of_interest:
            if isinstance(item, str):
                if (item in ALL) and (item not in roi):
                    roi.append(item)
    else:
        raise STDFError("objects_from_indexed_file(%s, %index, %records_of_interest) : Unsupported records_of_interest" % (FileName, index, records_of_interest))
    for REC_ID in roi:
        if REC_ID in index:
            for fp in index[REC_ID]:
                OBJ = create_record_object(version, endian, REC_ID, get_record_from_file_at_position(fd, fp, RLF))
                yield OBJ


def get_record_from_file_at_position(fd, offset, REC_LEN_FMT):
     fd.seek(offset)
     header = fd.read(4)
     REC_LEN = struct.unpack(REC_LEN_FMT, header[:2])[0]
     footer = fd.read(REC_LEN)
     return header+footer

def read_record(fd, RHF):
     '''
     This method will read one record from fd (at the current fp) with record header format RHF, and return the raw record
     '''
     header = fd.read(4)
     REC_LEN, REC_TYP, REC_SUB = struct.unpack(RHF, header)
     footer = fd.read(REC_LEN)
     return REC_LEN, REC_TYP, REC_SUB, header+footer

def read_indexed_record(fd, fp, RHF):
     fd.seek(fp)
     header = fd.read(4)
     REC_LEN, REC_TYP, REC_SUB = struct.unpack(RHF, header)
     footer = fd.read(REC_LEN)
     return REC_LEN, REC_TYP, REC_SUB, header+footer

    
class records_from_file(object):
    '''
    Generator class to run over the records in FileName.
    The return values are 4-fold : REC_LEN, REC_TYP, REC_SUB and REC
    REC is the complete record (including REC_LEN, REC_TYP & REC_SUB)
    if unpack indicates if REC is to be the raw record or the unpacked object.
    of_interest can be a list of records to return. By default of_interest is void
    meaning all records (of FileName's STDF Version) are used.
    '''
    debug = False

    def __init__(self, FileName, unpack=False, of_interest=None):
        if self.debug:
            print("initializing 'records_from_file")
        if isinstance(FileName, str):
            self.keep_open = False
            if not os.path.exists(FileName):
                raise STDFError("'%s' does not exist" %(FileName))
#           seimit : adding support for compressed files
            compression = extension_from_magic_number_in_file(FileName)
            if compression[0] == '.xz':
                import lzma
                self.fd = lzma.open(FileName, 'rb')
                self.parse_FAR()
            elif compression[0] == '.bz2':
                import bz2
                self.fd = bz2.open(FileName, 'rb')
                self.parse_FAR()
            elif compression[0] == '.gz':
                import gzip
                self.fd = gzip.open(FileName, 'rb')
                self.parse_FAR()
            elif compression[0] == '.zip':
                import zipfile
                zfile = zipfile.ZipFile(FileName, 'r')
                for name in zfile.namelist():
                    self.fd = zfile.open(name)
                    self.parse_FAR()
            else:
                # Assume standard binary stdf file
                self.endian = get_STDF_setup_from_file(FileName)[0]
                self.version = 'V%s' % struct.unpack(
                    'B', get_bytes_from_file(FileName, 5, 1))
                self.fd = open(FileName, 'rb')
        elif isinstance(FileName, io.IOBase):
            self.keep_open = True
            self.fd = FileName
            self.parse_FAR()
        else:
            raise STDFError("'%s' is not a string or an open file descriptor" %(FileName))
        self.unpack = unpack
        self.fmt = '%sHBB' % self.endian
        TS2ID = ts_to_id(self.version)
        if of_interest == None:
            self.records_of_interest = TS2ID
        elif isinstance(of_interest, list):
            ID2TS = id_to_ts(self.version)
            tmp_list = []
            for item in of_interest:
                if isinstance(item, str):
                    if item in ID2TS:
                        if ID2TS[item] not in tmp_list:
                            tmp_list.append(ID2TS[item])
                elif isinstance(item, tuple) and len(item) == 2:
                    if item in TS2ID:
                        if item not in tmp_list:
                            tmp_list.append(item)
            self.records_of_interest = tmp_list
        else:
            raise STDFError("objects_from_file(%s, %s) : Unsupported of_interest" % (
                FileName, of_interest))

    def parse_FAR(self):
        ptr = self.fd.tell()
        self.fd.seek(4)
        buff = self.fd.read(2)
        CPU_TYPE, STDF_VER = struct.unpack('BB', buff)
        if CPU_TYPE == 1:
            self.endian = '>'
        elif CPU_TYPE == 2:
            self.endian = '<'
        else:
            self.endian = '?'
        self.version = 'V%s' % STDF_VER
        self.fd.seek(ptr)


    def __del__(self):
        if not self.keep_open:
            self.fd.close()

    def __iter__(self):
        return self

    def __next__(self):
        while self.fd != None:
            header = self.fd.read(4)
            if len(header) != 4:
                raise StopIteration
            else:
                REC_LEN, REC_TYP, REC_SUB = struct.unpack(self.fmt, header)
                footer = self.fd.read(REC_LEN)
                if (REC_TYP, REC_SUB) in self.records_of_interest:
                    if len(footer) != REC_LEN:
                        raise StopIteration()
                    else:
                        return create_record_object(self.version, self.endian, (REC_TYP, REC_SUB), header+footer)
#                       seimit: the old code, does not corresponds to the README file
#                               but it is used in STDFFile class
#                        if self.unpack:
#                            return REC_LEN, REC_TYP, REC_SUB, create_record_object(self.version, self.endian, (REC_TYP, REC_SUB), header+footer)
#                        else:
#                            return REC_LEN, REC_TYP, REC_SUB, header+footer
#
#                       The code from the README will fail always because the code
#                       returns a tuple (REC_LEN, REC_TYP, REC_SUB, header+footer):
#                           for REC in STDF.records_from_file(f.name):
#                               print(REC.to_atdf())
#
#                       The correct README example according the code must be
#                           for REC in STDF.records_from_file(f.name, unpack=True):
#                               print(REC[3].to_atdf())
#
#                       To fix issue #18 I changed the code, just to return the 
#                       record's object which will fix the issue.

def create_record_object(Version, Endian, REC_ID, REC=None):
    '''
    This function will create and return the appropriate Object for REC
    based on REC_ID. REC_ID can be a 2-element tuple or a string.
    If REC is not None, then the record will also be unpacked.
    
    ToDo : 
        records from STDF v.3 must be removed
        records from STDF v.4 extension must be comment out for now
    '''
    retval = None
    REC_TYP = -1
    REC_SUB = -1
    if Version not in supported().versions():
        raise STDFError("Unsupported STDF Version : %s" % Version)
    if Endian not in ['<', '>']:
        raise STDFError("Unsupported Endian : '%s'" % Endian)
    if isinstance(REC_ID, tuple) and len(REC_ID) == 2:
        TS2ID = ts_to_id(Version)
        if (REC_ID[0], REC_ID[1]) in TS2ID:
            REC_TYP = REC_ID[0]
            REC_SUB = REC_ID[1]
            REC_ID = TS2ID[(REC_TYP, REC_SUB)]
    elif isinstance(REC_ID, str):
        ID2TS = id_to_ts(Version)
        if REC_ID in ID2TS:
            (REC_TYP, REC_SUB) = ID2TS[REC_ID]
    else:
        raise STDFError("Unsupported REC_ID : %s" % REC_ID)

    if REC_TYP != -1 and REC_SUB != -1:
        retval = create_record(Version, Endian, REC_ID, REC)
    return retval

def create_record(Version, Endian, REC_ID, REC):
    
    retval = None
    
    if REC_ID == 'ATR':
        retval = ATR(Version, Endian, REC)
    elif REC_ID == 'BPS':
        retval = BPS(Version, Endian, REC)
    elif REC_ID == 'CDR':
        retval = CDR(Version, Endian, REC)
    elif REC_ID == 'CNR':
        retval = CNR(Version, Endian, REC)
    elif REC_ID == 'DTR':
        retval = DTR(Version, Endian, REC)
    elif REC_ID == 'EPS':
        retval = EPS(Version, Endian, REC)
    elif REC_ID == 'FAR':
        retval = FAR(Version, Endian, REC)
    elif REC_ID == 'FTR':
        retval = FTR(Version, Endian, REC)
    elif REC_ID == 'GDR':
        retval = GDR(Version, Endian, REC)
    elif REC_ID == 'HBR':
        retval = HBR(Version, Endian, REC)
    elif REC_ID == 'MIR':
        retval = MIR(Version, Endian, REC)
    elif REC_ID == 'MPR':
        retval = MPR(Version, Endian, REC)
    elif REC_ID == 'MRR':
        retval = MRR(Version, Endian, REC)
    elif REC_ID == 'NMR':
        retval = NMR(Version, Endian, REC)
    elif REC_ID == 'PCR':
        retval = PCR(Version, Endian, REC)
    elif REC_ID == 'PGR':
        retval = PGR(Version, Endian, REC)
    elif REC_ID == 'PIR':
        retval = PIR(Version, Endian, REC)
    elif REC_ID == 'PLR':
        retval = PLR(Version, Endian, REC)
    elif REC_ID == 'PMR':
        retval = PMR(Version, Endian, REC)
    elif REC_ID == 'PRR':
        retval = PRR(Version, Endian, REC)
    elif REC_ID == 'PSR':
        retval = PSR(Version, Endian, REC)
    elif REC_ID == 'PTR':
        retval = PTR(Version, Endian, REC)
    elif REC_ID == 'RDR':
        retval = RDR(Version, Endian, REC)
    elif REC_ID == 'SBR':
        retval = SBR(Version, Endian, REC)
    elif REC_ID == 'SDR':
        retval = SDR(Version, Endian, REC)
    elif REC_ID == 'SSR':
        retval = SSR(Version, Endian, REC)
    elif REC_ID == 'STR':
        retval = STR(Version, Endian, REC)
    elif REC_ID == 'TSR':
        retval = TSR(Version, Endian, REC)
    elif REC_ID == 'VUR':
        retval = VUR(Version, Endian, REC)
    elif REC_ID == 'WCR':
        retval = WCR(Version, Endian, REC)
    elif REC_ID == 'WIR':
        retval = WIR(Version, Endian, REC)
    elif REC_ID == 'WRR':
        retval = WRR(Version, Endian, REC)
    return retval

def dict_to_rec(rec_dict, Endian):
    '''
    Creates a STDF record based on prevoiusly exported dictionary from STDF record
    Input : dictionary with STDF records in format field_name : field_value

    Returns
    -------
    STDF record object.

    '''
    Version = 'V4'
    if Endian not in ['<', '>']:
        raise STDFError("Unsupported Endian : '%s'" % Endian)

    rec_id = rec_dict["rec_id"]
    record = create_record(Version, Endian, rec_id, None)

    for k, v in rec_dict.items():
        if v != None and k != 'rec_id':
            record.set_value(k,v)
    record._update_rec_len()
    return record      

# The following code was comment out and located in records.py->STDR.py and moved here
# class xrecords_from_file(object):
#     '''
#     This is a *FAST* iterator class that returns the next record from an STDF file each time it is called.
#     It is fast because it doesn't check versions, extensions and it doesn't unpack the record and skips unknown records.
#     '''
#
#     def __init__(self, FileName, of_interest=None):
#         #TODO: add a record_list of records to return
#         if isinstance(FileName, str):
#             try:
#                 stdf_file = File(FileName)
#             except:
#                 raise StopIteration
#             self.fd = stdf_file.open()
#         elif isinstance(FileName, File):
#             stdf_file = FileName
#             self.fd = FileName.open()
#         else:
#             raise STDFError("records_from_file(%s) : Unsupported 'FileName'" % FileName)
#         self.endian = stdf_file.endian
#         self.version = stdf_file.version
#         TS2ID = ts_to_id(self.version)
#         if of_interest==None:
#             self.of_interest = list(TS2ID.keys())
#         elif isinstance(of_interest, list):
#             ID2TS = id_to_ts(self.version)
#             tmp_list = []
#             for item in of_interest:
#                 if isinstance(item, str):
#                     if item in ID2TS:
#                         if ID2TS[item] not in tmp_list:
#                             tmp_list.append(ID2TS[item])
#                 elif isinstance(item, tuple) and len(item)==2:
#                     if item in TS2ID:
#                         if item not in tmp_list:
#                             tmp_list.append(item)
#             self.of_interest = tmp_list
#         else:
#             raise STDFError("records_from_file(%s, %s) : Unsupported of_interest" % (FileName, of_interest))
#
#     def __del__(self):
#         self.fd.close()
#
#     def __iter__(self):
#         return self
#
#     def __next__(self):
#         while self.fd!=None:
#             while True:
#                 header = self.fd.read(4)
#                 if len(header)!=4:
#                     raise StopIteration
#                 REC_LEN, REC_TYP, REC_SUB = struct.unpack('HBB', header)
#                 footer = self.fd.read(REC_LEN)
#                 if len(footer)!=REC_LEN:
#                     raise StopIteration
#                 if (REC_TYP, REC_SUB) in self.of_interest:
#                     return REC_LEN, REC_TYP, REC_SUB, header+footer
#
# class xobjects_from_file(object):
#     '''
#     This is an iterator class that returns the next object (unpacked) from an STDF file.
#     It will take care of versions and extensions, and unrecognized records will simply be skipped.
#     '''
#     def __init__(self, FileName, of_interest=None):
#         if isinstance(FileName, str):
#             try:
#                 stdf_file = File(FileName)
#             except:
#                 raise STDFError("objects_from_file(%s, %s) : File doesn't exist" % (FileName, of_interest))
#             self.fd = stdf_file.open()
#         elif isinstance(FileName, File):
#             self.fd = FileName.open()
#         else:
#             raise STDFError("objects_from_file(%s) : Unsupported 'FileName'" % FileName)
#         self.endian = stdf_file.endian
#         self.version = stdf_file.version
#         TS2ID = ts_to_id(self.version)
#         if of_interest==None:
#             of_interest = TS2ID
#         elif isinstance(of_interest, list):
#             ID2TS = id_to_ts(self.version)
#             tmp_list = []
#             for item in of_interest:
#                 if isinstance(item, str):
#                     if item in ID2TS:
#                         if ID2TS[item] not in tmp_list:
#                             tmp_list.append(ID2TS[item])
#                 elif isinstance(item, tuple) and len(item)==2:
#                     if item in TS2ID:
#                         if item not in tmp_list:
#                             tmp_list.append(item)
#             of_interest = tmp_list
#         else:
#             raise STDFError("objects_from_file(%s, %s) : Unsupported of_interest" % (FileName, of_interest))
#         self.of_interest = of_interest
#         self.fmt = '%sHBB' % self.endian
#
#     def __del__(self):
#         self.fd.close()
#
#     def __iter__(self):
#         return self
#
#     def __next__(self):
#         while True:
#             header = self.fd.read(4)
#             if len(header)!=4:
#                 raise StopIteration
#             else:
#                 REC_LEN, REC_TYP, REC_SUB = struct.unpack(self.fmt, header)
#                 footer = self.fd.read(REC_LEN)
#                 if len(footer)!=REC_LEN:
#                     raise StopIteration
#                 else:
#                     record = header + footer
#                     if (REC_TYP, REC_SUB) in self.of_interest:
#                         recobj = create_record_object(self.version, self.endian, (REC_TYP, REC_SUB), record)
#                         return (recobj)


# class open(object):
#     '''
#     file opener that opens an STDF file transparently (gzipped or not)
#     '''
#     def __init__(self, fname):
#         f = open(fname)
#         # Read magic number (the first 2 bytes) and rewind.
#         magic_number = f.read(2)
#         f.seek(0)
#         # Encapsulated 'self.f' is a file or a GzipFile.
#         if magic_number == '\x1f\x8b':
#             self.f = gzip.GzipFile(fileobj=f)
#         else:
#             self.f = f
#
#         # Define '__enter__' and '__exit__' to use in 'with' blocks.
#         def __enter__(self):
#             return self
#         def __exit__(self, type, value, traceback):
#             try:
#                 self.f.fileobj.close()
#             except AttributeError:
#                 pass
#             finally:
#                 self.f.close()
#
#         # Reproduce the interface of an open file by encapsulation.
#         def __getattr__(self, name):
#             return getattr(self.f, name)
#         def __iter__(self):
#             return iter(self.f)
#         def next(self):
#             return next(self.f)



# def wafer_map(data, parameter=None):
#     '''
#     data is a pandas data frame, it has at least 5 columns ('X_COORD', 'Y_COORD', 'LOT_ID', 'WAFER_ID' and the parameter)
#     If the parameter is not named the following order of 'parameters' will be used :
#         'HARD_BIN'
#         'SOFT_BIN'
#         'PART_PF'

#     '''
#     pass

# def get_MIR_from_file(FileName):
#     '''
#     This function will just get the MIR (near the start of the file) from the FileName and return it.
#     it must *NOT* be guaranteed that FileName exists or is an STDF File.
#     '''
#     endian, version = get_STDF_setup_from_file(FileName)
#     mir = None
#     if endian!=None and version!=None: # file exists and is an STDF file
#         for record in xrecords_from_file(FileName):
#             _, REC_TYP, REC_SUB, REC = record
#             if (REC_TYP, REC_SUB) == (1, 10):
#                 mir = MIR(version, endian, REC)
#                 break
#     return mir

# def get_partcount_from_file(FileName):
#     '''
#     This function will return the number of parts contained in FileName.
#     it must *NOT* be guaranteed that FileName exists or is an STDF File.
#     '''

# def save_STDF_index(FileName, index):
#     '''
#     '''
#     if os.path.exists(FileName) and os.path.isfile(FileName):
#         Path, Name = os.path.split(FileName)
#         Base, Ext = os.path.splitext(Name)
#         if Ext in ['.stdf', '.pbz2']:
#             pickle_file = os.path.join(Path, "%s.pbz2" % Base)
#         else:
#             raise Exception("FileName should have '.stdf' or '.pbz2' extension")
#         with bz2.open(pickle_file, 'wb') as fd:
#             pickle.dump(index, fd)
#     else:
#         raise Exception("File {} does not exists or is not a file!".format(FileName))




# -*- coding: utf-8 -*-
from . import STDR

class MMR(STDR):
     def __init__(self, version=None, endian=None, record = None):
         self.id = 'MMR'
         self.local_debug = False
         if version==None or version=='V4':
             self.version = 'V4'
             self.info=    '''
 Memory Model Record (V4, Memory:2010.1)
 ---------------------------------------

 Function:
     This record is used to store the memory model information in STDF.
     The record allows storing the logic level information of the model.
     It does not have any fields to store the physical information except height and width.
     The physical information can be optionally linked to the record through a reference to the file.

 Frequency:
     Once per memory model.

 Location:
     It can occur after all the Instance Description Records(IDRs) and before any Frame Specification Records (FSRs),
     Bit Stream Specification Records (BSRs) and any Test specific records e.g. Parametric Test Record (PTR),
     Functional Test Record (FTRs), Scan Test Record (STR) and Memory Test Record (MTRs).
 '''
             self.fields = {
                 'REC_LEN'  : {'#' :  0, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                 'REC_TYP'  : {'#' :  1, 'Type' : 'U*1',  'Ref' :       None, 'Value' :    1, 'Text' : 'Record type                           ', 'Missing' : None},
                 'REC_SUB'  : {'#' :  2, 'Type' : 'U*1',  'Ref' :       None, 'Value' :   95, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                 'ASR_IDX'  : {'#' :  3, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Unique identifier for this ASR record ', 'Missing' : None},
                 'STRT_IDX' : {'#' :  4, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Cycle Start index flag                ', 'Missing' : None},
                 'ALGO_CNT' : {'#' :  5, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'count (k) of Algorithms descriptions  ', 'Missing' :    0},
                 'ALGO_NAM' : {'#' :  6, 'Type' : 'xC*n', 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Names Name of the Algorithm  ', 'Missing' :   []},
                 'ALGO_LEN' : {'#' :  7, 'Type' : 'xC*n', 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Complexity of algorithm      ', 'Missing' :   []},
                 'FILE_ID'  : {'#' :  8, 'Type' : 'xC*n', 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Name of the file with descr. ', 'Missing' :   []},
                 'CYC_BGN'  : {'#' :  9, 'Type' : 'xU*8', 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Starting cycle number        ', 'Missing' :   []},
                 'CYC_END'  : {'#' : 10, 'Type' : 'xU*8', 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of End Cycle number             ', 'Missing' :   []}
             }
         else:
             raise STDR.STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
         self._default_init(endian, record)




# -*- coding: utf-8 -*-
from . import STDR

class ASR(STDR):
     def __init__(self, version=None, endian=None, record=None):
         self.id = ''
         self.local_debug = False
         # Version
         if version==None or version=='V4':
             self.version = 'V4'
             self.info=    '''
 Algorithm Specification Record (V4, Memory:2010.1)
 --------------------------------------------------

 Function:
     This record is used to store the algorithms that are applied during a memory test. Table 11 Algorithm Specification Record (ASR) Record

 Frequency:
     * Once per unique memory test specification.

 Location:
     It can occur after all the Memory Model Records(MMRs) and before any Test specific records
     e.g. Parametric Test Record (PTR), Functional Test Record (FTRs), Scan Test Record (STR) and Memory Test Record (MTRs).
 '''
             self.fields = {
                 'REC_LEN'  : {'#' :  0, 'Type' : 'U*2'  , 'Ref' :       None, 'Value' : None, 'Text' : 'Bytes of data following header                      ', 'Missing' : None, 'Note' : ''},
                 'REC_TYP'  : {'#' :  1, 'Type' : 'U*1'  , 'Ref' :       None, 'Value' :    0, 'Text' : 'Record type                                         ', 'Missing' : None, 'Note' : ''},
                 'REC_SUB'  : {'#' :  2, 'Type' : 'U*1'  , 'Ref' :       None, 'Value' :   20, 'Text' : 'Record sub-type                                     ', 'Missing' : None, 'Note' : ''},
                 'ASR_IDX'  : {'#' :  3, 'Type' : 'U*2'  , 'Ref' :       None, 'Value' : None, 'Text' : 'Unique identifier for this ASR record               ', 'Missing' :    0, 'Note' : ''},
                 'STRT_IDX' : {'#' :  4, 'Type' : 'U*1'  , 'Ref' :       None, 'Value' : None, 'Text' : 'Cycle Start index flag                              ', 'Missing' :    0, 'Note' : ''},
                 'ALGO_CNT' : {'#' :  5, 'Type' : 'U*1'  , 'Ref' :       None, 'Value' : None, 'Text' : 'count (k) of Algorithms descriptions                ', 'Missing' :    0, 'Note' : ''},
                 'ALGO_NAM' : {'#' :  6, 'Type' : 'xC*n' , 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Names of the Algorithms                    ', 'Missing' :   [], 'Note' : ''},
                 'ALGO_LEN' : {'#' :  7, 'Type' : 'xC*n' , 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Complexity of algorithm  (e.g., 13N)       ', 'Missing' :   [], 'Note' : ''},
                 'FILE_ID'  : {'#' :  8, 'Type' : 'xC*n' , 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Name of the file with algorithm description', 'Missing' :   [], 'Note' : ''},
                 'CYC_BGN'  : {'#' :  9, 'Type' : 'xU*8' , 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of Starting cycle number for the Algorithms   ', 'Missing' :   [], 'Note' : ''},
                 'CYC_END'  : {'#' : 10, 'Type' : 'xU*8' , 'Ref' : 'ALGO_CNT', 'Value' : None, 'Text' : 'Array of End Cycle number for the algorithm         ', 'Missing' :   [], 'Note' : ''}
             }
         else:
             raise STDR.STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
         self._default_init(endian, record)



'''
Created on Jun 14, 2014

@author: $Author: tho $

This Library implements Date and Time manipulations in a convenient way.

License : GPL
'''
import datetime
import os
import time

from PyQt5.QtCore import QDate
from PyQt5.QtCore import QDateTime
from PyQt5.QtCore import QTime


def is_date(stamp):
    '''
    This function will return True or False, depending if the supplied stamp
    can be interpreted as a date string of format DDMMYYYY
    '''
    if type(stamp) == str:
        if len(stamp) == 8:
            if stamp.isdigit():
                DD = int(stamp[:2])
                MM = int(stamp[2:4])
                YYYY = int(stamp[4:])
                if DD > 31 : return False
                if MM > 12 : return False
                return True
            else:
                return False
        else:
            return False
    else:
        return False


def is_datecode(stamp):
    '''
    This function will return True or False, depending if the supplied stamp
    can be interpreted as a date-code string of the format YYWWD
    '''
    if type(stamp) == str:
        if len(stamp) == 5:
            year = stamp[0:2]
            if not year.isdigit():
                return False
            week = stamp[2:4]
            if week.isdigit():
                week = int(week)
                if (week < 1) or (week > 53):
                    return False
            else:
                return False
            day = stamp[4]
            if day.isdigit():
                day = int(day)
                if (day < 1) or (day >7):
                    return False
                else:
                    return True
            else:
                return False
        else:
            return False
    else:
        return False

def iso_year_start(iso_year):
    "The gregorian calendar date of the first day of the given ISO year"
    fourth_jan = datetime.date(iso_year, 1, 4)
    delta = datetime.timedelta(fourth_jan.isoweekday()-1)
    return fourth_jan - delta

def iso_to_gregorian(iso_year, iso_week, iso_day):
    "Gregorian calendar date for the given ISO year, week and day"
    year_start = iso_year_start(iso_year)
    return year_start + datetime.timedelta(days=iso_day-1, weeks=iso_week-1)

def url_datetime_string_to_epoch(url_datetime):
    retval = url_datetime
    print(retval)
    retval = retval.split(',')[1].strip()
    retval = retval.replace('GMT', '').strip()
    retval = datetime.datetime.strptime(retval, "%d %b %Y %H:%M:%S")
    return int((retval - datetime.datetime(1970,1,1)).total_seconds())

def micronas_time_structure_to_epoch(Datum=0, Uhrzeit=0):
    '''
    Micronas uses the following legacy structures:

    Datum : YYYYYYMM MMDDDDD0 (2 bytes)
                YYYYYY --> Year  6 bit, range = 0..63, offset is 2000
                MMMM ----> Month 4 bit, range = 0..15, 0=january .. 11=december
                DDDDD ---> Day   5 bit, range = 0..31, 0 is not used, 1=first day of the month and so on

    Uhrzeit : HHHHHHHH MMMMMMMM (2 bytes)
                HHHHHHHH --> Hour    8 bit, range = 0..255 only 0..23 is used
                MMMMMMMM --> Minutes 8 bit, range = 0..255 only 0..59 is used

    This function returns the seconds since epoch for the given Datum and Uhrzeit as an initializer for the DT objects
    '''
    year = (Datum & (0b111111 << 10)) >> 10
    month = (Datum & (0b1111 << 6)) >> 6
    day_of_month = (Datum & (0b11111 << 1)) >> 1
    hours = (Uhrzeit & (0b11111111 << 8)) >> 8
    minutes = Uhrzeit & 0b11111111
    dt = DT("%2s%2s%4s" % (day_of_month, month+1, year+2000))
    return dt.epoch+(hours*60*60)+(minutes*60)

class DTOerror(RuntimeError):
    '''
    module exception
    '''
    def __init__(self, arg):
        self.args = arg

    def __str__(self):
        return ''.join(self.args)

    def __repr__(self):
        return ''.join(self.args)


class TD(object):
    '''
    Time Differnce Object
    '''
    def __init__(self, diff):
        self.diff = diff

    def seconds(self):
        return self.diff

    def minutes(self):
        return self.diff/60

    def hours(self):
        return self.diff/(60*60)

    def days(self):
        return self.diff/(60*60*24)

    def weeks(self):
        return self.diff/(60*60*24*7)

    def months(self):
        return self.diff/(60*60*24*365/12)

    def years(self):
        return self.diff/(60*60*24*365)

    def __str__(self):
        '''
        3 years, 2 months, 3 weeks, 4 days 15 hours 20 min 3 seconds
        '''
        retval = ''
        if self.years()>=1:

            retval+="%d year"

        return retval

class DT(object):
    '''
    Date and Time Object
    '''
    def __init__(self, stamp=None):
        loct = int(time.time())
        self.tz = (time.mktime(time.gmtime(loct)) - loct) / 3600
        self.dst = time.localtime()[8]
        self.__call__(stamp)

    def __call__(self, stamp=None):
        if stamp == None:
            self.epoch = int(time.time())
        elif isinstance(stamp, int) or isinstance(stamp, float):
            self.epoch = int(stamp)
        elif isinstance(stamp, str):
            if is_datecode(stamp): # YYWWD
                # set epoch to the beginning of the datecode
                # http://stackoverflow.com/questions/5882405/get-date-from-iso-week-number-in-python
                iso_year = int(stamp[:2])
                if iso_year > 70: # epoch = 1 Jan 1970 ;-)
                    iso_year += 1900
                else:
                    iso_year += 2000
                iso_week = int(stamp[2:4])
                iso_day = int(stamp[4])
                temp = iso_to_gregorian(iso_year, iso_week, iso_day)
                self.epoch = int((datetime.datetime(temp.year,temp.month,temp.day) - datetime.datetime(1970,1,1)).total_seconds())
            elif is_date(stamp): # DDMMYYYY
                DD = int(stamp[:2])
                MM = int(stamp[2:4])
                YYYY = int(stamp[4:])
                self.epoch = int((datetime.datetime(YYYY, MM, DD) - datetime.datetime(1970,1,1)).total_seconds())
            elif os.path.exists(stamp): # file path
                self.epoch =  os.stat(stamp)[6]
            else:
                raise Exception("can not interprete the string '%s' as an initializer " % stamp)
        elif isinstance(stamp, QDateTime):
            self.epoch = stamp.currentSecsSinceEpoch()
        elif isinstance(stamp, QDate):
            DD = stamp.dayOfWeek()
            MM = stamp.month()
            YYYY = stamp.year()
            self.epoch = int((datetime.datetime(YYYY, MM, DD) - datetime.datetime(1970,1,1)).total_seconds())
        else:
            raise Exception("Don't now how to handle type(%s)=%s" % (stamp, type(stamp)))
        self._populate()

    def _populate(self):
        t = time.gmtime(self.epoch)
        d = datetime.date(t.tm_year, t.tm_mon, t.tm_mday)
        self.datecode = "%4d%02d%1d" % (d.isocalendar()[0], d.isocalendar()[1], d.isocalendar()[2]) # YYYYWWD
        self.datecode = self.datecode[2:] # YYWWD
        self.min = t.tm_min # minute of the hour [0 .. 59]
        if self.min < 15:
            self.qhour = 1
        elif self.min < 30:
            self.qhour = 2
        elif self.min < 45:
            self.qhour = 3
        else:
            self.qhour = 4
        self.hour = t.tm_hour # hour of the day [0 .. 23]
        self.sec = t.tm_sec # seconds of the min [0..59]
        self.wday = d.isocalendar()[2]
        if self.wday == 1:
            self.wday_name = 'Monday'
        elif self.wday == 2:
            self.wday_name = 'Tuesday'
        elif self.wday == 3:
            self.wday_name = 'Wednesday'
        elif self.wday == 4:
            self.wday_name = 'Thursday'
        elif self.wday == 5:
            self.wday_name = 'Friday'
        elif self.wday == 6:
            self.wday_name = 'Saturday'
        elif self.wday == 7:
            self.wday_name = 'Sunday'
        else: # should never reach this point
            raise DTOerror
        self.mday = t.tm_mday # day of the month [1 .. 31]
        self.yday = t.tm_yday # day of the year [1 .. 365/366] depending on leap year
        self.week = d.isocalendar()[1] # week of the year [1 .. 52/53] aka 'Work Week'
        self.month = t.tm_mon # month of the year [1 .. 12]
        if self.month == 1:
            self.month_name = 'January'
        elif self.month == 2:
            self.month_name = 'February'
        elif self.month == 3:
            self.month_name = 'March'
        elif self.month == 4:
            self.month_name = 'April'
        elif self.month == 5:
            self.month_name = 'May'
        elif self.month == 6:
            self.month_name = 'June'
        elif self.month == 7:
            self.month_name = 'July'
        elif self.month == 8:
            self.month_name = 'August'
        elif self.month == 9:
            self.month_name = 'September'
        elif self.month == 10:
            self.month_name = 'October'
        elif self.month == 11:
            self.month_name = 'November'
        elif self.month == 12:
            self.month_name = 'December'
        else: # should never reach this point
            raise DTOerror
        if self.month in [1, 2, 3]: self.quarter = 1
        elif self.month in [4, 5, 6]: self.quarter = 2
        elif self.month in [7, 8, 9]: self.quarter = 3
        else: self.quarter = 4
        self.year = t.tm_year # year
        self.datetime = datetime.datetime(self.year, self.month, self.mday, self.hour, self.min, self.sec)
        self.date = datetime.date(self.year, self.month, self.mday)
        self.time = datetime.time(self.hour, self.min, self.sec)
        utc_offset = (-self.tz*3600)+(self.dst*3600)
        self.QDateTime = QDateTime()
        self.QDateTime.setOffsetFromUtc(int(utc_offset)) # timezone + day light saving
        self.QDateTime.setSecsSinceEpoch(self.epoch)
        self.QDate = self.QDateTime.date()
        self.QTime = self.QDateTime.time()

    def boh(self):
        '''
        Returns the epoch (is thus gmt based) for the Begin Of the Hour from the underlaying epoch.
        The epoch of the object remains unchanged.
        '''
        return int((datetime.datetime(self.year, self.month, self.mday, self.hour) - datetime.datetime(1970,1,1)).total_seconds())

    def eoh(self):
        '''
        Returns the epoch (is thus gmt based) for the End Of the Hour from the underlaying epoch.
        The underlaying epoch of the object remains unchanged.
        '''
        return self.boh()+3600-1

    def bod(self):
        '''
        Returns the epoch (is thus gmt based) for the Begin Of the Day from the underlaying epoch.
        The epoch of the object remains unchanged.
        '''
        return int((datetime.datetime(self.year, self.month, self.mday) - datetime.datetime(1970,1,1)).total_seconds())

    def eod(self):
        '''
        Returns the epoch (is thus gmt based) for the End Of the Day from the underlaying epoch.
        The underlaying epoch of the object remains unchanged.
        '''
        return self.bod() + (24*60*60)

    def bow(self):
        '''
        Returns the epoch (is thus gmt based) for the Begin Of the Week from the underlaying epoch.
        The underlaying epoch of the object remains unchanged.
        '''
        temp = iso_to_gregorian(self.year, self.week, 1)
        return int((datetime.datetime(temp.year,temp.month,temp.day) - datetime.datetime(1970,1,1)).total_seconds())

    def eow(self):
        '''
        Returns the epoch (is thus gmt based) for the End Of the Week from the underlaying epoch.
        The underlaying epoch of the object remains unchanged.
        '''
        return self.bow() + (7*24*60*60)

    def bom(self):
        '''
        Returns the epoch (is thus gmt based) of the Begin Of the Month from the undelaying epoch.
        The underlying epoch of the object remains unchanged.
        '''
        return DT("01%2s%4s" % (self.month, self.year))

    def eom(self):
        '''
        Returns the epoch (is thus gmt based) of the End Of the Month from the undelaying epoch.
        The underlying epoch of the object remains unchanged.
        '''
        if self.month == 12:
            return DT("0101%4s" % (self.year + 1)).epoch -1
        else:
            return DT("01%2s%4s" % (self.month + 1, self.year)).epoch - 1

    def boy(self):
        '''
        Returns the epoch (is thus gmt based) of the Begin Of the Year from the underlaying epoch.
        The underlying epoch of the object remains unchanged.
        '''
        return DT("0101%4s" % self.year).epoch

    def eoy(self):
        '''
        Returns the epoch (is thus gmt based) of the End Of the Year from the undelaying epoch.
        The undelaying epoch of the object remains unchanged.
        '''
        return DT("0101%4s" % (self.year + 1)).epoch -1

    def local(self):
        '''
        Returns the epoch (is thus gmt based) for the LOCAL time.
        The underlaying epoch remains unchanged.
        '''
        return self.epoch - (self.tz * 3600) + (abs(self.dst) * 3600)

    def __sub__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return TD(self.epoch - other.epoch)

        return NotImplemented


    def __lt__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.epoch < other.epoch:
            return True
        else:
            return False

    def __le__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.epoch <= other.epoch:
            return True
        else:
            return False

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.epoch == other.epoch:
            return True
        else:
            return False

    def __ne__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.epoch != other.epoch:
            return True
        else:
            return False

    def __gt__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.epoch > other.epoch:
            return True
        else:
            return False

    def __ge__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.epoch >= other.epoch:
            return True
        else:
            return False

    def __repr__(self):
        #Monday, January 6 2014 @ 13:22:11 (Q1 2014)
        return "%s, %s %s %s @ %02d:%02d:%02d (Q%s %s)" % (self.wday_name, self.month_name, self.mday, self.year, self.hour, self.min, self.sec, self.quarter, self.datecode)

class ChronoMeter(object):

    def __init__(self):
        self.reset()

    def reset(self):
        self.lap = 1
        self.scoreboard = {}

    def start(self):
        self.TimeStart = time.time()

    def stop(self):
        self.TimeStop = time.time()
        self.scoreboard[self.lap] = (self.TimeStart, self.TimeStop - self.TimeStart)
        self.lap =+ 1

    def laps(self):
        return self.Laps

    def time(self):
        return self.TotalTime

    def laptime(self, lap):
        return self.scoreboard[lap][1]

    def averagelaptime(self):
        pass #TODO: Implement

    def __repr__(self):
        print(self.__str__())

    def __str__(self):
        return "Timed %f sec for %d laps (%f sec per lap)" % (self.time(), self.laps(), self.laptime())

if __name__ == "__main__":
    now = DT()
    print("now (UTC) = %s" % now)
    print("now (local) = %s" % DT(now.local()))
    print("now tz = %s" % now.tz)
    print("now dst = %s" % now.dst)


    print("begin of week = %s" % DT(now.bow()))
    print("end of week = %s" % DT(now.eow()))
    print("begin of day = %s" % DT(now.bod()))
    print("end of day = %s" % DT(now.eod()))



# -*- coding: utf-8 -*-
from . import STDR

class PSR(STDR):
    def __init__(self, version=None, endian=None, record = None):

         self.id = 'PSR'
         self.local_debug = False
         if version==None or version=='V4':
             self.version = 'V4'
             self.info=    '''
# Pattern Sequence Record (V4-2007)
# ---------------------------------

# Function:
#     PSR record contains the information on the pattern profile for a specific executed scan test
#     as part of the Test Identification information. In particular it implements the Test Pattern
#     Map data object in the data model. It specifies how the patterns for that test were constructed.
#     There will be a PSR record for each scan test in a test program. A PSR is referenced by the STR
#     (Scan Test Record) using its PSR_INDX field

# Frequency:
#     ?!?

# Location:
#     ?!?
# '''
             self.fields = {
                 'REC_LEN'  : {'#' :  0, 'Type' : 'U*2', 'Ref' :       None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' :    None},
                 'REC_TYP'  : {'#' :  1, 'Type' : 'U*1', 'Ref' :       None, 'Value' :    1, 'Text' : 'Record type                           ', 'Missing' :    None},
                 'REC_SUB'  : {'#' :  2, 'Type' : 'U*1', 'Ref' :       None, 'Value' :   90, 'Text' : 'Record sub-type                       ', 'Missing' :    None},
                 'CONT_FLG' : {'#' :  3, 'Type' : 'B*1', 'Ref' :       None, 'Value' : None, 'Text' : 'PSR record(s) to follow if not 0      ', 'Missing' : ['0']*8},
                 'PSR_INDX' : {'#' :  4, 'Type' : 'U*2', 'Ref' :       None, 'Value' : None, 'Text' : 'PSR Record Index (used by STR records)', 'Missing' :    None},
                 'PSR_NAM'  : {'#' :  5, 'Type' : 'C*n', 'Ref' :       None, 'Value' : None, 'Text' : 'Symbolic name of PSR record           ', 'Missing' :      ''},
                 'OPT_FLG'  : {'#' :  6, 'Type' : 'B*1', 'Ref' :       None, 'Value' : None, 'Text' : 'Options Flag                          ', 'Missing' :    None},
                 'TOTP_CNT' : {'#' :  7, 'Type' : 'U*2', 'Ref' :       None, 'Value' : None, 'Text' : 'Count of sets in the complete data set', 'Missing' :       1},
                 'LOCP_CNT' : {'#' :  8, 'Type' : 'U*2', 'Ref' :       None, 'Value' : None, 'Text' : 'Count (k) of sets in this record      ', 'Missing' :       0},
                 'PAT_BGN'  : {'#' :  9, 'Type' :'xU*8', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : "Array of Cycle #'s patterns begins on ", 'Missing' :      []},
                 'PAT_END'  : {'#' : 10, 'Type' :'xU*8', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : "Array of Cycle #'s patterns stops at  ", 'Missing' :      []},
                 'PAT_FILE' : {'#' : 11, 'Type' :'xC*n', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : 'Array of Pattern File Names           ', 'Missing' :      []},
                 'PAT_LBL'  : {'#' : 12, 'Type' :'xC*n', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : 'Optional pattern symbolic name        ', 'Missing' :      []},
                 'FILE_UID' : {'#' : 13, 'Type' :'xC*n', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : 'Optional array of file identifier code', 'Missing' :      []},
                 'ATPG_DSC' : {'#' : 14, 'Type' :'xC*n', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : 'Optional array of ATPG information    ', 'Missing' :      []},
                 'SRC_ID'   : {'#' : 15, 'Type' :'xC*n', 'Ref' : 'LOCP_CNT', 'Value' : None, 'Text' : 'Optional array of PatternInSrcFileID  ', 'Missing' :      []}
             }
         else:
             raise STDR.STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
         self._default_init(endian, record)



# -*- coding: utf-8 -*-
from . import STDR

class CNR(STDR):
     def __init__(self, version=None, endian=None, record = None):
         self.id = 'CNR'
         self.local_debug = False
         if version==None or version=='V4':
             self.version = 'V4'
             self.info=    '''
# Cell Name Record (V4-2007)
# --------------------------

# Function:
#     This record is used to store the mapping from Chain and Bit position to the Cell/FlipFlop name.
#     A CNR record should be created for each Cell for which a name mapping is required.
#     Typical usage would be to create a record for each failing cell/FlipFlop.
#     A CNR with new mapping for a chain and bit position would override the previous mapping.

# Frequency:

# Location:
# '''
             self.fields = {
                 'REC_LEN'  : {'#' :  0, 'Type' : 'U*2' , 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                 'REC_TYP'  : {'#' :  1, 'Type' : 'U*1' , 'Ref' : None, 'Value' :    1, 'Text' : 'Record type                           ', 'Missing' : None},
                 'REC_SUB'  : {'#' :  2, 'Type' : 'U*1' , 'Ref' : None, 'Value' :   92, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                 'CHN_NUM'  : {'#' :  2, 'Type' : 'U*2' , 'Ref' : None, 'Value' : None, 'Text' : 'Chain number. (cfr STR:CHN_NUM)       ', 'Missing' :    0},
                 'BIT_POS'  : {'#' :  2, 'Type' : 'U*4' , 'Ref' : None, 'Value' : None, 'Text' : 'Bit position in the chain             ', 'Missing' :    0},
                 'CELL_NAM' : {'#' :  2, 'Type' : 'S*n' , 'Ref' : None, 'Value' : None, 'Text' : 'Scan Cell Name                        ', 'Missing' :   ''}
             }
         else:
             raise STDR.STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
         self._default_init(endian, record)

